#include "userprog/syscall.h"
#include <stdio.h>
#include <syscall-nr.h>
#include <userprog/process.h>
#include "threads/interrupt.h"
#include "threads/thread.h"
#include "devices/shutdown.h"
#include "filesys/filesys.h"
#include "filesys/file.h"


static void syscall_handler (struct intr_frame *);
void syscall_exit(int status);
void syscall_halt(void);
bool syscall_create(const char *, unsigned);
bool syscall_remove(const char*);
int syscall_open(const char *file);
void syscall_close(int fd);
tid_t syscall_exec(const char* cmd_line);

void
syscall_init (void) 
{
  intr_register_int (0x30, 3, INTR_ON, syscall_handler, "syscall");
}


static void
syscall_handler (struct intr_frame *f)
{
  printf ("system call!\n");
  int sys_code = *(int*)f->esp;
  int *ptr = f->esp;

  printf("System call number: %d\n", sys_code);

  switch(sys_code) {

    case SYS_EXIT:
      printf("Running SYS_EXIT\n");
      syscall_exit(*(ptr+1));
      break;

    case SYS_HALT:
      printf("Running SYS_HALT\n");
      syscall_halt();
      break;

    case SYS_CREATE:
      printf("Running SYS_CREATE\n");
      f->eax = syscall_create(*(ptr+4), *(ptr+5));
      break;

    case SYS_REMOVE:
      printf("Running SYS_REMOVE\n");
      f->eax = syscall_remove(*(ptr+1));
      break;

    case SYS_OPEN:
      printf("Running SYS_OPEN\n");
      int ret_open = syscall_open((const char*)*(ptr+1));
      f->eax = ret_open;
      break;

    case SYS_CLOSE:
      printf("Running SYS_CLOSE\n");
      syscall_close(*(ptr+1));
      break;

    case SYS_EXEC:
      printf("Running SYS_EXEC\n");
      f->eax = syscall_exec(*(ptr+1));
      break;

    default:
      printf("Unknown Syscall number\n");
      break;
  }

  thread_exit ();
}

/**************************************** 
    System Call Implementations
*****************************************/

/* Exits the current process */
void syscall_exit(int status)
{
  thread_exit();
} 

/* Shuts system down immediately */
void syscall_halt()
{
  shutdown_power_off();
}

/* Runs a given executable */
tid_t syscall_exec(const char* cmd_line)
{
  tid_t tid = process_execute(cmd_line);
  return tid;
}

/* Create file but does not open it */
bool syscall_create(const char * file, unsigned initial_size)
{
  if (file == NULL)
    return false;
    
  bool create_success = filesys_create(file, initial_size);
  printf("Creating file\n");
  printf("%s", create_success? "File Created\n":"File Not Created\n");
  return create_success;
}

/* Removes file from disk */
bool syscall_remove(const char* file)
{
  if (file == NULL)
    return false;
  bool remove_success = filesys_remove(file);
  return remove_success;
}

/* Opens given file */
int syscall_open(const char *file)
{
  struct file *data = filesys_open(file);
  if (data == NULL)
    return -1;
  int fd = 2;
  while (fd < 128)
  {
    if (thread_current()->fd_table[fd] == NULL)
    {
      thread_current()->fd_table[fd] = data;
      return fd;
      printf("Open system call\n");
    }
    return -1;
  }
}

/* Closes given file */
void syscall_close(int fd)
{
  struct thread *cur = thread_current();
  struct file *f = cur->fd_table[fd];
  printf("%s: close(%d)\n", cur->name, fd);
  file_close(f);
  printf("Close system call\n");
}